/*    nethackappview.cpp
 *
 * Nethack UI View class
 *
 * Copyright 2002-2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <gdi.h>
#include <eikenv.h>
#include <apgtask.h>
#include <eikspane.h>
#include <eikbtgpc.h>
#include "nethackappui.h"
#include "nethackappview.h"
#include "nethackui.hrh"

_LIT(KAssertPanic, "nethackappview.cpp");
#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KAssertPanic, __LINE__))

static const TInt KSmallFontWidth = 6;
static const TInt KSmallFontHeight = 8;

static const TInt KLargeFontWidth = 7;
static const TInt KLargeFontHeight = 14;

static const TInt KDefaultTermWidth = 80;
static const TInt KDefaultTermHeight = 24;


CNethackAppView::CNethackAppView(CNethackAppUi *aAppUi) {
    iAppUi = aAppUi;
}

CNethackAppView::~CNethackAppView() {
    delete iTerminal;
}


void CNethackAppView::ConstructL(const TRect &aRect) {
    CreateWindowL();
    SetRect(aRect);

    // Determine the correct size and position for the terminal. We'll aim
    // at a default size initially, instead of making it as big as possible
    TInt termWidth = Rect().Width();
    TInt termHeight = Rect().Height();
    if ( termWidth > (KDefaultTermWidth * KSmallFontWidth) ) {
        termWidth = KDefaultTermWidth * KSmallFontWidth;
    }
    if ( termHeight > (KDefaultTermHeight * KSmallFontHeight) ) {
        termHeight = KDefaultTermHeight * KSmallFontHeight;
    }
    TInt termX = Rect().iTl.iX + (Rect().Width() - termWidth) / 2;
    TInt termY = Rect().iTl.iY + (Rect().Height() - termHeight) / 2;
    iTermRect = TRect(TPoint(termX, termY), TSize(termWidth, termHeight));
    
    iTerminal = new (ELeave) CTerminalControl();
    iTerminal->ConstructL(iTermRect, Window());
    ActivateL();
}


void CNethackAppView::Draw(const TRect & /*aRect*/) const {

    CWindowGc &gc = SystemGc();
    gc.Reset();
    gc.SetClippingRect(Rect());

    // Determine terminal window borders and draw a rectangle around it
    TRect borderRect = iTermRect;
    borderRect.Grow(1, 1);
    gc.DrawRect(borderRect);

    // Clear everything outside the terminal
    TRegionFix<5> clearReg(Rect());
    clearReg.SubRect(borderRect);
    assert(!clearReg.CheckError());
    const TRect *rects = clearReg.RectangleList();
    TInt numRects = clearReg.Count();
    
    gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
    gc.SetBrushColor(KRgbWhite);
    gc.SetPenStyle(CGraphicsContext::ENullPen);
    while ( numRects-- ) {
        gc.DrawRect(*(rects++));
    }
}


TInt CNethackAppView::CountComponentControls() const {
    return 1;
}


CCoeControl *CNethackAppView::ComponentControl(TInt aIndex) const {
    
    switch ( aIndex ) {
        case 0:
            return iTerminal;

        default:
            assert(EFalse);
    }

    return NULL;
}


TKeyResponse CNethackAppView::OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                          TEventCode aType) {

    // Normally we'll just let the terminal control handle the event, but
    // there are some exceptions:
    // - If the user presses Chr without other keys,
    //   we'll launch the character selection dialog.
    if ( (aType == EEventKeyDown) && (aKeyEvent.iCode == EKeyNull) &&
         ((aKeyEvent.iModifiers & EModifierLeftFunc) != 0) ) {
        iChrDown = ETrue;
    } else if ( iChrDown &&
                (aType == EEventKeyUp) &&
                (aKeyEvent.iCode == EKeyNull) &&
                ((aKeyEvent.iModifiers & EModifierLeftFunc) == 0) ) {
        // Chr up/down without other keys
        iChrDown = EFalse;
//        iAppUi->HandleCommandL(ECmdSendSpecialCharacter);
        return EKeyWasConsumed;
    } else {
        iChrDown = EFalse;
    }

    return iTerminal->OfferKeyEventL(aKeyEvent, aType);
}


CTerminalControl *CNethackAppView::Terminal() {
    return iTerminal;
}


void CNethackAppView::SetFontL(TBool aLargeFont) {

    // Set font to the terminal control
    iLargeFont = aLargeFont;
    iTerminal->SetFontL(iLargeFont);

    ResizeTerminal();
}


void CNethackAppView::ResizeTerminal() {

    if ( !iTerminal ) {
        return;
    }
    
    // Determine the terminal window size
    TInt fontWidth = KSmallFontWidth;
    TInt fontHeight = KSmallFontHeight;
    if ( iLargeFont ) {
        fontWidth = KLargeFontWidth;
        fontHeight = KLargeFontHeight;
    }

    TInt termWidth = Rect().Width();
    TInt termHeight = Rect().Height();

    if ( (!iFullScreen) && (!iLargeFont) ) {
        // Small font and non-fullscreen: use default size
        if ( termWidth > (KDefaultTermWidth * fontWidth) ) {
            termWidth = KDefaultTermWidth * fontWidth;
        }
        if ( termHeight > (KDefaultTermHeight * fontHeight) ) {
            termHeight = KDefaultTermHeight * fontHeight;
        }
        
    }

    // Round size down to the largest possible terminal that contains whole
    // characters
    termWidth = fontWidth * (termWidth / fontWidth);
    termHeight = fontHeight * (termHeight / fontHeight);
    assert((termWidth > 0) && (termHeight > 0));

    // Set terminal size and position
    TInt termX = Rect().iTl.iX + (Rect().Width() - termWidth) / 2;
    TInt termY = Rect().iTl.iY + (Rect().Height() - termHeight) / 2;
    iTermRect = TRect(TPoint(termX, termY), TSize(termWidth, termHeight));
    iTerminal->SetRect(iTermRect);
    DrawDeferred();
}


void CNethackAppView::SetFullScreenL(TBool aFullScreen) {

    CEikonEnv *eikonEnv = CEikonEnv::Static();

    iFullScreen = aFullScreen;
    
    if ( iFullScreen ) {
        // Hide indicator bar and CBA:
        eikonEnv->AppUiFactory()->StatusPane()->MakeVisible(EFalse);
        eikonEnv->AppUiFactory()->ToolBar()->MakeVisible(EFalse);

        // Set view to cover full screen
        SetRect(iAppUi->ClientRect());
        
    } else {
        // Show indicator bar and CBA:
        eikonEnv->AppUiFactory()->StatusPane()->MakeVisible(ETrue);
        eikonEnv->AppUiFactory()->ToolBar()->MakeVisible(ETrue);

        // Set view to cover the remaining area
        SetRect(iAppUi->ClientRect());
    }

    DrawDeferred();
}


void CNethackAppView::SizeChanged() {
    ResizeTerminal();
}
