/*    terminalcontrol.cpp
 *
 * A terminal UI control class
 *
 * Copyright 2002-2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <gdi.h>
#include <eikenv.h>
#include <e32svr.h>
#include "terminalcontrol.h"

_LIT(KPanic, "TerminalControl");

_LIT(KTerminalFont, "Terminal");
_LIT(KTerminalControl, "terminalcontrol.cpp");
static const TInt KSmallFontHeight = 8;
static const TInt KLargeFontHeight = 14;
static const TInt KDefaultLargeFont = EFalse;

#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KTerminalControl, __LINE__))

// Small: 6x8
// Large: 7x14
// The bold version is one pixel wider, and thus unusable!


CTerminalControl::CTerminalControl() {
    iGrayed = EFalse;
}


CTerminalControl::~CTerminalControl() {

    if ( iFont != NULL ) {
        CEikonEnv::Static()->ScreenDevice()->ReleaseFont(iFont);
        iFont = NULL;
    }
    delete [] iChars;
    delete [] iAttributes;
}


void CTerminalControl::ConstructL(const TRect &aRect,
                                  RWindow &aContainerWindow) {

    SetContainerWindowL(aContainerWindow);
    
    // Set the default font
    SetFontL(KDefaultLargeFont);

    SetRect(aRect);
    Resize();
    AllocateBuffersL();
    Clear();
}


void CTerminalControl::SetObserver(MTerminalObserver *aObserver) {
    iObserver = aObserver;
}


void CTerminalControl::RemoveObserver() {
    iObserver = NULL;
}


void CTerminalControl::SetFontL(TBool aLargeFont) {

    TInt fontHeight = KSmallFontHeight;    
    if ( aLargeFont ) {
        fontHeight = KLargeFontHeight;
    }

    // Release old fonts
    if ( iFont != NULL ) {
        CEikonEnv::Static()->ScreenDevice()->ReleaseFont(iFont);
        iFont = NULL;
    }
    
    // Get the font
    CGraphicsDevice *dev = CEikonEnv::Static()->ScreenDevice();    
    TInt twips = dev->VerticalPixelsToTwips(fontHeight);
    TFontSpec spec(KTerminalFont, twips);
    User::LeaveIfError(dev->GetNearestFontInTwips(iFont, spec));

    // Get font info and resize terminal
    iFontHeight = iFont->HeightInPixels();
    iFontWidth = iFont->WidthZeroInPixels();
}


void CTerminalControl::SetGrayed(TBool aGrayed) {
    iGrayed = aGrayed;
    DrawDeferred();
}


// Resizes the terminal in response to size or font change
void CTerminalControl::Resize() {

    TRect rect = Rect();
    TInt width = rect.iBr.iX - rect.iTl.iX;
    TInt height = rect.iBr.iY - rect.iTl.iY;

    assert((iFontWidth > 0) && (iFontHeight > 0));
    assert((width >= iFontWidth) && (height >= iFontHeight));
    iCharWidth = width / iFontWidth;
    iCharHeight = height / iFontHeight;
    if ( iObserver )
        iObserver->TerminalSizeChanged(iCharWidth, iCharHeight);
    TRAPD(error, AllocateBuffersL());
    if ( error != KErrNone ) {
        User::Panic(KPanic, error);
    }
}


// (re)allocate character and attribute buffers
void CTerminalControl::AllocateBuffersL() {
    delete [] iChars;
    iChars = NULL;
    delete [] iAttributes;
    iAttributes = NULL;

    iChars = new (ELeave) TText[iCharWidth*iCharHeight];
    iAttributes = new (ELeave) TTerminalAttribute[iCharWidth*iCharHeight];

    Clear();
}


// Clear the buffers
void CTerminalControl::Clear() {
    TText *c = iChars;
    TTerminalAttribute *a = iAttributes;
    TInt num = iCharWidth * iCharHeight;
    while ( num-- ) {
        *c++ = ' ';
        a->iFgColor = KRgbBlack;
        a->iBgColor = KRgbWhite;
        a->iBold = EFalse;
        a->iUnderline = EFalse;
        a++;
    }
}


// Draws text on the terminal window. The coordinates are zero-based
// character coordinates inside the terminal.
void CTerminalControl::DrawText(TInt aX, TInt aY, const TDesC &aText,
                                TBool aBold, TBool aUnderline,
                                TRgb aForeground, TRgb aBackground) {
    // Don't draw text while grayed out
    if ( iGrayed ) {
        return;
    }

    // Check that we are at least partially on screen
    if ( (aX < 0) || (aY < 0) || (aX >= iCharWidth) || (aY >= iCharHeight) ) {
        return;
    }

    // Write the text to our character and attribute buffers
    TInt numChars = aText.Length();
    if ( numChars > (iCharWidth - aX) ) {
        numChars = iCharWidth - aX;
    }
    TText *c = &iChars[aX + aY*iCharWidth];
    TTerminalAttribute *a = &iAttributes[aX + aY * iCharWidth];
    for ( TInt i = 0; i < numChars; i++ ) {
        *c++ = aText[i];
        a->iFgColor = aForeground;
        a->iBgColor = aBackground;
        a->iBold = aBold;
        a->iUnderline = aUnderline;
        a++;
    }
    
    // Actually draw the text on screen
    CWindowGc &gc = SystemGc();
    gc.Activate(Window());
    gc.UseFont(iFont);

    DoDraw(gc, aX, aY, aText.Left(numChars), aBold, aUnderline, aForeground,
           aBackground);

    gc.DiscardFont();
    gc.Deactivate();
}


// Actually draw text on screen
void CTerminalControl::DoDraw(CGraphicsContext &aGc, TInt aX, TInt aY,
                              const TDesC &aText,
                              TBool /*aBold*/, TBool aUnderline,
                              TRgb aForeground, TRgb aBackground) const {

    // We'll draw the text inside a filled box. This should avoid flickering
    // compared to filling first and drawing afterwards

    // Brush is used for background
    aGc.SetBrushColor(aBackground);
    aGc.SetBrushStyle(CGraphicsContext::ESolidBrush);

    // Pen is used for foreground
    aGc.SetPenColor(aForeground);
    aGc.SetPenStyle(CGraphicsContext::ESolidPen);

    // Handle underline. Sadly we cannot use a bold version of the Terminal
    // font to do real bold, since it is one pixel wider than the normal
    // version
    if ( aUnderline ) {
        aGc.SetUnderlineStyle(EUnderlineOn);
    }    

    // Draw the text inside a box
    TInt ascent = iFont->AscentInPixels();
    TInt x = Rect().iTl.iX;
    TInt y = Rect().iTl.iY;
    TRect textRect(x + aX * iFontWidth, y + aY * iFontHeight,
                   x + (aX + aText.Length()) * iFontWidth,
                   y + (aY+1) * iFontHeight);
    aGc.DrawText(aText, textRect, ascent);

    // Reset underline attribute
    if ( aUnderline ) {
        aGc.SetUnderlineStyle(EUnderlineOff);
    }
}


// CCoeControl::OfferKeyEventL()
TKeyResponse CTerminalControl::OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                              TEventCode aType) {

    if ( aType == EEventKey ) {
        TUint code = aKeyEvent.iCode;

        // Pass keypress to the observer
        if ( iObserver )
            iObserver->KeyPressed((TKeyCode)code, aKeyEvent.iModifiers);
        return EKeyWasConsumed;
    }

    return EKeyWasNotConsumed;
}


// CCoeControl::Draw()
void CTerminalControl::Draw(const TRect & /*aRect*/) const {

    CWindowGc &gc = SystemGc();
    gc.Reset();

    if ( iGrayed ) {
        gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
        gc.SetBrushColor(TRgb(0xcccccc));
        gc.SetPenStyle(CGraphicsContext::ENullPen);
        gc.DrawRect(Rect());
    }
    else {
        gc.UseFont(iFont);

        // We'll find the longest possible run of identical attributes on
        // each line and draw those at one go
        for ( TInt y = 0; y < iCharHeight; y++ ) {
            TText *lChars = &iChars[y * iCharWidth];
            TTerminalAttribute *lAttribs = &iAttributes[y * iCharWidth];
            TInt x = 0;

            while ( x < iCharWidth ) {
                TTerminalAttribute a0 = lAttribs[x];
                TTerminalAttribute *a = &lAttribs[x];
                TInt num = 0;
                while ( ((x+num) < iCharWidth) &&
                        (a->iFgColor == a0.iFgColor) &&
                        (a->iBgColor == a0.iBgColor) &&
                        (a->iBold == a0.iBold) &&
                        (a->iUnderline == a0.iUnderline) ) {
                    a++;
                    num++;
                }
                DoDraw(gc, x, y, TPtrC(&lChars[x], num), a0.iBold,
                       a0.iUnderline, a0.iFgColor, a0.iBgColor);
                x += num;
            }
        }

        gc.DiscardFont();        
    }
}


// CCoeControl::SizeChanged()
void CTerminalControl::SizeChanged() {
    Resize();
}


// CCoeControl::InputCapabilities()
TCoeInputCapabilities CTerminalControl::InputCapabilities() const {
    return TCoeInputCapabilities(TCoeInputCapabilities::EAllText |
                                 TCoeInputCapabilities::ENavigation);
}
